/**************************************************************************************************

Copyright (c) 2014   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan
Copyright (c) 2007-2012   Tsunehiko Kato

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mtk_satellites.h"
#include  "mtk_satellite_data.h"
#include  "mtk_mode.h"
#include  "mtk_objects.h"
#include  "../common/astro/astro_const.h"
#include  "mtk_gl_common.h"
#include  "mtk_OpenGL.h"
#include  "mtk_draw_common.h"
#include  "mtk_solar_system.h"
#include  "mtk_sun.h"
#include  "mtk_planets.h"
#include  "mtk_planet_position.h"
#include  "mtk_planet_orientation.h"
#include  "mtk_planet_phys_params.h"
#include  "mtk_planet_eclipse.h"
#include  "mtk_place_name_earth_mars.h"
#include  "mtk_planetarium.h"
#include  "mtk_target.h"
#include  "mtk_label.h"
#include  "mtk_switches.h"
#include  "mtk_colors.h"
#include  "mtk_calc_alpha.h"
#include  "mtk_dialog_box.h"
#include  "mtk_font.h"
#include  <vector>





TopoSphere	tpMoon;	// n`



//-----------
//  Externs
//-----------
extern bool bDrawLowRes;  //****
extern float TopoZoomRate[5];
extern char TopoZoomMode;
vector2d	get_landing_pos();





//-------------
//  Text size
//-------------
static float  SatelliteNameSize;



//------------------
//  Landing radius
//------------------
double
get_landing_radius_satellite(int idx)
{
	double Rx, Ry, Rz;
	get_satellite_size( MODE_3D, idx, false, Rx, Ry, Rz);
	return  Rx * _1km_pc;
}




//----------------
//  Prepare draw
//----------------
bool
prepare_draw_satellite(const mtk_draw_struct& ds, int idx, double* pRb, double* pRh)
{
	//---  Visible
	if  ( !test_disp_switch(DISPSW_SATELLITE,	DISPM_DISP) )  return false;


	//---  Radius
	double Rx, Ry, Rz;
	get_satellite_size(ds.tyMode, idx, false, Rx, Ry, Rz);
	*pRb = *pRh = Rx * _1km_pc;

	return  true;
}



//---------
//  Scale
//---------
double
get_scale_satellite(int idx)
{
	double Rx, Ry, Rz;
	get_satellite_size(MODE_3D, idx, false, Rx, Ry, Rz);
	return  30.0 * Rx * _1km_pc;
}



//-------------------
//   Visible radius
//-------------------
double
calc_visible_radius_satellite(int idx)
{
	return  get_satellite_position_relative_to_planet(idx).norm();
}



//---------
//  Label
//---------
bool
is_satellite_label_visible(int idx)
{
	if ( !test_disp_switch(DISPSW_SATELLITE, DISPM_DISP) || !test_disp_switch(DISPSW_SATELLITE, DISPM_NAME ) )  return  false;
	if ( test_disp_switch(DISPSW_SATELLITE, DISPM_SELECT) && !is_satellite_selected(idx))  return  false;

	return  true;
}


float
get_satellite_label_alpha(const mtk_draw_struct& ds, int idx)
{
	float scl = 1.5f;
	const vector3d  rpos = get_satellite_position( idx ) - ds.camera.get_pos();
	float  Rv = scl * calc_visible_radius_satellite( idx );
	float  Dth = 2.0f * Rv / rpos.norm();

	return  calc_alpha_by_apparent_diameter( Dth );
}







//-----------------
//  Load Textures
//-----------------
void
load_satellites_textures()
{
	SetSplashText("q̃eNX`Ǎ...", "Loading textures of satellites...");
	load_satellites_textures_sub(TexShrinkRate);
}





//-------------
//  Text size
//-------------
void
set_satellites_name_size(float sz)
{
	SatelliteNameSize = sz;
}

float
get_satellites_name_size()
{
	return  SatelliteNameSize;
}



//----------
//  Orbits
//----------
void
register_satellites_orbit_curves()
{
	const int N = get_satellites_num();

	const float ow = get_orbit_width();
	for (int i=0; i<N; i++) {
		space_curve  sc( mtk_object(OBJ_TYPE_SATELLITE, i, 0.0f, false), ow, MTK_COLOR_ORBIT_SATELLITE );
		register_space_curve( sc );
	}
}


void
set_satellites_orbit_curve(space_curve& sc, double t)
{
	const int idx = sc.get_related_object().get_index();
	const int N = sc.get_num_points();

	for (int i=0; i<N; i++) {
		double phi = 2.0 * M_PI * double(i) / N;
		vector3d pos = calc_satellite_orbit_pos(idx, t, phi);
		sc.set_point(i, pos);
	}
}

void
make_satellites_orbit_curve(space_curve& sc, double t, int div_N)
{
	sc.set_num_points( div_N );
	set_satellites_orbit_curve(sc, t);
}


float
get_satellite_orbit_alpha(const mtk_draw_struct& ds, int idx)
{
	int pid = get_satellite_planet_index(idx);
	float Re = get_planet_Re(ds.tyMode, pid) * _1km_pc;
	const float dscl = 8 * Re;
	float  alp = calc_alpha_by_disapear_scale_inside( ds.scale, dscl );

	return  alp * get_satellite_label_alpha( ds, idx );
}


//-----------
//  ȉq
//-----------
//***********************
bool
is_satellite_selected(int idx)
{
	switch (idx) {
		case SAT_MOON:
		case SAT_PHOBOS:	case SAT_DEIMOS:
		case SAT_IO:		case SAT_EUROPA:	case SAT_GANYMEDE:	case SAT_CALLISTO:	case SAT_AMALTHEA:	case SAT_HIMALIA:
		case SAT_MIMAS:		case SAT_ENCELADUS:	case SAT_TETHYS:	case SAT_DIONE:		case SAT_RHEA:		case SAT_TITAN:		case SAT_IAPETUS:
		case SAT_ARIEL:		case SAT_UMBRIEL:	case SAT_TITANIA:	case SAT_OBERON:	case SAT_MIRANDA:
		case SAT_TRITON:	case SAT_NEREID:
		case SAT_CHARON:
			return  true;
		default:	break;
	}

	return  false;
}



void
load_moon_topograph(directory& dir, const char* filename)
{
	tpMoon.load(dir.get_path(filename));
	tpMoon.set_elv_fac(TopoZoomRate[TopoZoomMode]);
}





//--------
//  Draw
//--------
void
set_satellite_orientation(int idx, const double& W_shift)
{
	set_object_orientation( get_satellite_sat_to_ec_matrix(idx) );
	glRotated(W_shift,0,0,1);
}

void
draw_satellite(const mtk_draw_struct& ds, int idx, const vector3d& rpos, double scl)
{
	vector3d r_sat = get_satellite_position(idx);		// q̈ʒuiSW, in pcj

	double Rx, Ry, Rz;
	get_satellite_size(ds.tyMode, idx, false, Rx, Ry, Rz);
	Rx *= scl * _1km_pc;
	Ry *= scl * _1km_pc;
	Rz *= scl * _1km_pc;


	const frustum frs = ds.get_frustum();
	bool bWire = false;

	glPushMatrix();
		glTranslatef(rpos.x, rpos.y, rpos.z);

		if (!bSideLight)	set_light_from_sun(r_sat);
		else				set_side_light(ds);

		set_satellite_orientation(idx, 180.0);		// ]щ]pZbg


		if ( test_disp_switch(DISPSW_PLANET, DISPM_AXIS) ) draw_planet_axis(Rz);


		char tex_no = get_satellite_texture_index(idx);
		if (tex_no != -1) {
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, tex_no);
		} else {
			glDisable(GL_TEXTURE_2D);
		}


		if (idx == SAT_MOON) {
			const float screen_w = get_screen_w();
			const float screen_z = get_screen_z();
			const mtk_object&  tgt = get_target();
			bool bTarget = (tgt.get_type() == OBJ_TYPE_SATELLITE && tgt.get_index() == idx);

			glPushMatrix();

			glScalef(Rx, Ry, Rz);
			glLineWidth(1.0f);
			glColor3f(0, 1, 0);


			tpMoon.draw(frs, bWire, bDrawLowRes);

			if (test_disp_switch(DISPSW_PLANET, DISPM_NAME) && ds.tySubMode != SUBMODE_3DCHART) {
				wgl_font &fnt = get_font();
				PlaceName& pn = get_place_name(PLACE_OBJ_MOON);
				bool bHorizontalMode = (get_landing_mode() == LANDING_MODE_HORIZONTAL);
				pn.draw_on_TopoSphere(tpMoon, screen_w, screen_z, get_font(), (ds.lang == LANG_ENG), ds.ltf, bHorizontalMode);
			}

			
			if (bSurfacePointer && bTarget && ds.tySubMode != SUBMODE_3DCHART) {
				glDisable(GL_LIGHTING);
				glDisable(GL_TEXTURE_2D);
				glEnable(GL_BLEND);
				vector2d lon_lat = get_landing_pos();
				draw_pointer_on_TopoSphere(tpMoon, screen_w, screen_z, get_font(), lon_lat.x, lon_lat.y, ds.ltf);
			}
			


			glPopMatrix();
		}
		else {
			draw_celestial_object(Rx, Ry, Rz);
		}



		if (swEclipse & (ECLIPSE_DISP | ECLIPSE_BORDER)) {
			if (idx == SAT_MOON) {
				glPushMatrix();
					glScalef(Rx,Ry,Rz);
					draw_satellite_eclipse(ds, SAT_MOON, PLNT_EARTH);
				glPopMatrix();
			}
		}
	glPopMatrix();
}



void
draw_a_satellite(const mtk_draw_struct& ds, int idx, const vector3d& rpos, const double& scl)
{
	glEnable(GL_LIGHT0);
	float mat_d[] = {1,1,1,1}, mat_a[] = {0.5,0.5,0.5,1};
	mat_a[0] = mat_a[1] = mat_a[2] = get_lighting_ambient();
	glMaterialfv( GL_FRONT_AND_BACK , GL_DIFFUSE,   mat_d );
	glMaterialfv( GL_FRONT_AND_BACK , GL_AMBIENT,   mat_a );

	draw_satellite(ds, idx, rpos, scl);

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
}





//------------
//  Hit-test
//------------
bool
hit_test_satellite(const mtk_draw_struct& ds, int idx, const vector3d& rpos, const vector3d& nv)
{
	matrix3d  M_E2SAT;
	vector3d  _rpos, _nv;
	M_E2SAT = get_satellite_ec_to_sat_matrix(idx);
	_rpos = M_E2SAT * rpos;
	_nv = M_E2SAT * nv;

	// G
	double Rx, Ry, Rz;
	get_satellite_size(MODE_3D, idx, false, Rx, Ry, Rz);
	if (Rx <= 0.0)  return  false;
	Rx *= _1km_pc;
	Ry *= _1km_pc;
	Rz *= _1km_pc;

	double fy, fz;
	fy = Ry / Rx;
	fz = Rz / Rx;
	_rpos.y /= fy;
	_rpos.z /= fz;
	_nv.y /= fy;
	_nv.z /= fz;
	_nv = normalize(_nv);

	// _
	double R, b, c, D, t;
	R = Rx;
	b = dot_product(_rpos, _nv);
	c = _rpos.norm2() - R*R;
	D = b*b - c;

	if (D <= 0.0)  return false;
	t = -b -sqrt(D);
	if (t < 0.0)  return  false;

	return  true;
}




//--------
//  Info
//--------
void
make_object_info_satellite(const mtk_draw_struct& ds, int idx, std::vector<std::string>& text)
{
	bilingual bszName = get_satellite_name(idx);

	if (ds.lang == LANG_ENG) {
		text.push_back( bszName.text_E );
		text.push_back( "Type: Satellite" );
	} else {
		text.push_back( bszName.text_J );
		text.push_back( "ށFq" );
	}
}







//---------------
//  Planetarium
//---------------
void
draw_a_satellite_sub(const float& R, int idx)
{
	glEnable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glEnable(GL_NORMALIZE);

	glColor3f(1,1,1);

	char tex_no = get_satellite_texture_index(idx);
	if (tex_no != -1) {
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, tex_no);
	} else {
		glDisable(GL_TEXTURE_2D);
	}

	glScalef(R, R, R);
	gl_draw_LOD_sphere();

	glDisable(GL_NORMALIZE);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
}



void
draw_a_satellite_planetarium(const mtk_draw_struct& ds, int idx)
{
	const bilingual bszName = get_satellite_name(idx);
	if (ds.tgt_ID == get_object_ID( bszName.text_E ))  return;


	vector3d  spos = get_satellite_position(idx);


	//---  eʒuƃTCY  ---
	vector3d  rpos = spos - ds.camera.get_pos();
	double  L = rpos.norm();
	// ʒu
	const double R0 = 1.002;

	vector3d pos = R0 * rpos/L;
	// a
	double Rx, Ry, Rz, Rs;
	get_satellite_size(ds.tyMode, idx, false, Rx, Ry, Rz);
	Rs = Rx / L * _1km_pc;




	// p
	double th = Rs/pos.norm();

	th *= get_planetarium_zooming_factor();

	const double thc1 = 3*ds.rad_per_pixel;
	const double thc2 = 0.1*thc1;
	const double thc3 = 0.01*thc1;
	if (th < thc3)  return;

	bool bPoint;
	float af;
	if (th < thc1) {
		bPoint = true;
		af = th/thc2;
		if (af > 1) af = 1;
	} else {
		bPoint = false;
		af = 1.0f;
	}


	//---  Draw  ---
	glEnable(GL_DEPTH_TEST);	//****

	glPushMatrix();
		glTranslatef(pos.x, pos.y, pos.z);
	
		if (!bPoint) {
			//---  TCYŕ`  ---
			glPushMatrix();
				set_light_from_sun( spos );
				set_satellite_orientation(idx, 180);
				draw_a_satellite_sub(Rs, idx);
				if ( idx==SAT_MOON && (swEclipse & (ECLIPSE_DISP | ECLIPSE_BORDER)) ) {
					int  pidx = get_satellite_planet_index(idx);
					draw_satellite_eclipse(ds, idx, pidx);
				}
			glPopMatrix();
		} else {
			//---  _ŕ`  ---
			glDisable(GL_LIGHTING);
			glDisable(GL_BLEND);
			glDisable(GL_TEXTURE_2D);
			glPointSize(1.5);
			glBegin(GL_POINTS);
				glColor3f(af,af,af);
				glVertex3f(0,0,0);
			glEnd();
		}

		// O
		if ( test_disp_switch(DISPSW_SATELLITE, DISPM_NAME) ) {
			float sz;
			int col_idx;

			if (idx == SAT_MOON) {
				sz = get_special_object_name_size();
				col_idx = MTK_COLOR_NAME_SPECIAL_OBJECT;
			} else {
				sz = SatelliteNameSize;
				col_idx = MTK_COLOR_NAME_SPECIAL_OBJECT;
			}

			prepare_3Dtext_1(Rs*1.41*ds.plnt_R, 45, sz*ds.ltf, 1.0f, false);
			
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_LIGHTING);
			glEnable(GL_BLEND);

			const RGBf& col = af * get_color(col_idx);
			glColor3f(col.r, col.g, col.b);
			draw_bilingual(bszName);
		}
	glPopMatrix();
}


void
draw_satellites_planetarium(const mtk_draw_struct& ds)
{
	float mat_d[] = {1,1,1,1};
	float mat_a[] = {0.5,0.5,0.5,1};
	mat_a[0] = mat_a[1] = mat_a[2] = get_lighting_ambient();
	glMaterialfv( GL_FRONT_AND_BACK , GL_DIFFUSE,   mat_d );
	glMaterialfv( GL_FRONT_AND_BACK , GL_AMBIENT,   mat_a );

	glEnable(GL_LIGHT0);

	for (int i=0; i<get_satellites_num(); i++) {
		draw_a_satellite_planetarium(ds, i);
	}
}





